#| -*-Scheme-*-

Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
    1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
    2017, 2018, 2019 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301,
USA.

|#

;;;; R7RS libraries: loader
;;; package: (runtime library loader)

(declare (usual-integrations))

;;;; Syntax

(define (syntax-r7rs-source source db)
  (register-r7rs-source! source (copy-library-db db))
  (r7rs-source->scode-file source))

(define-automatic-property '(contents imports-used)
    '(parsed-contents imports exports imports-environment)
  #f
  (lambda (contents imports exports env)
    (receive (body bound free)
	(syntax-library-forms (expand-contents contents) env)
      (let ((imports-to (map library-import-to imports)))
	(if (not (lset<= eq? free imports-to))
	    (warn "Library has free references not provided by imports:"
		  (lset-difference eq? free imports-to)))
	(let ((exports-from (map library-export-from exports))
	      (bound* (lset-union eq? bound imports-to)))
	  (if (not (lset<= eq? exports-from bound*))
	      (warn "Library export refers to unbound identifiers:"
		    (lset-difference eq? exports-from bound*)))))
      (values body
	      (filter (lambda (import)
			(memq (library-import-to import) free))
		      imports)))))

(define (expand-contents contents)
  (append-map (lambda (directive)
		(case (car directive)
		  ((include)
		   (parameterize ((param:reader-fold-case? #f))
		     (append-map read-file
				 (cdr directive))))
		  ((include-ci)
		   (parameterize ((param:reader-fold-case? #t))
		     (append-map read-file
				 (cdr directive))))
		  ((begin)
		   (cdr directive))
		  (else
		   (error "Unknown content directive:" directive))))
	      contents))

;;;; Imports environment

(define (import-environment-available? import db)
  (let ((name (library-import-from-library import)))
    (and (registered-library? name db)
	 ((registered-library name db) 'has? 'environment))))

(define (make-environment-from-imports imports db)
  (let ((env
	 (make-root-top-level-environment
	  (delete-duplicates (map library-import-to imports) eq?))))
    (add-imports-to-env! imports env db #f)
    env))

(define (add-imports-to-env! imports env db allow-conflicts?)
  (for-each
   (lambda (group)
     (let ((exporter
	    (library-exporter
	     (registered-library
	      (library-import-from-library (car group))
	      db))))
       (if (not allow-conflicts?)
	   (let ((conflicts
		  (let ((bindings (environment-bindings env)))
		    (filter (lambda (import)
			      (let ((value
				     (exporter (library-import-from import)))
				    (name (library-import-to import)))
				(let ((b (assq name bindings)))
				  (and b
				       (pair? (cdr b))
				       (not
					(values-equivalent? value (cadr b)))))))
			    group))))
	     (if (pair? conflicts)
		 (error "Conflcting imports:" conflicts))))
       (for-each (lambda (import)
		   (let ((value (exporter (library-import-from import)))
			 (name (library-import-to import)))
		     (cond ((macro-reference-trap? value)
			    (environment-define-macro
			     env name
			     (macro-reference-trap-transformer value)))
			   ((unassigned-reference-trap? value)
			    ;; nothing to do
			    )
			   (else
			    (environment-define env name value)))))
		 group)))
   (group-imports-by-source imports)))

(define (group-imports-by-source imports)
  (let ((table (make-equal-hash-table)))
    (for-each (lambda (import)
		(hash-table-update! table
				    (library-import-from-library import)
				    (lambda (imports)
				      (cons import imports))
				    (lambda ()
				      '())))
	      imports)
    (hash-table-values table)))

(define (values-equivalent? v1 v2)
  (cond ((unassigned-reference-trap? v1)
	 (unassigned-reference-trap? v2))
	((macro-reference-trap? v1)
	 (and (macro-reference-trap? v2)
	      (eqv? (macro-reference-trap-transformer v1)
		    (macro-reference-trap-transformer v2))))
	(else (eqv? v1 v2))))

(define-automatic-property 'imports-environment '(imports db)
  (lambda (imports db)
    (every (lambda (import)
	     (import-environment-available? import db))
	   imports))
  make-environment-from-imports)

(define (environment . import-sets)
  (let ((db (current-library-db)))
    (make-environment-from-imports (import-sets->imports import-sets db)
				   db)))

(define (import! import-set #!optional env)
  (let ((db (current-library-db)))
    (add-imports-to-env! (import-sets->imports (list import-set) db)
			 (if (default-object? env)
			     (nearest-repl/environment)
			     (guarantee environment? env 'import!))
			 db
			 #t)))

(define (import-sets->imports import-sets db)
  (let ((imports (expand-parsed-imports (map parse-import-set import-sets) db)))
    (maybe-load-libraries! imports db)
    (let ((unavailable
	   (remove (lambda (import)
		     (import-environment-available? import db))
		   imports)))
      (if (pair? unavailable)
	  (error "Imported libraries unavailable:"
		 (library-imports-from unavailable))))
    imports))

(define (scheme-report-environment version)
  (if (not (eqv? version 5))
      (error "Unsupported version:" version))
  (environment '(scheme r5rs)))

(define (null-environment version)
  (if (not (eqv? version 5))
      (error "Unsupported version:" version))
  (environment '(only (scheme r5rs)
		      ... => _ and begin case cond define define-syntax delay do
		      else if lambda let let* let-syntax letrec letrec-syntax or
		      quasiquote quote set! syntax-rules)))

;;;; Evaluation

(define (eval-r7rs-source source db)
  (let ((program (register-r7rs-source! source db)))
    (if program
	(library-eval-result program))))

(define (eval-r7rs-scode-file scode pathname db)
  (let ((libraries
	 (let ((filename (->namestring pathname)))
	   (map (lambda (library)
		  (scode-library->library library filename))
		(r7rs-scode-file-elements scode)))))
    (register-libraries! libraries db)
    (let loop ((libraries libraries) (result unspecific))
      (if (pair? libraries)
	  (loop (cdr libraries)
		(library-eval-result (car libraries)))
	  result))))

(define-automatic-property '(eval-result environment)
    '(contents imports-environment name)
  #f
  (lambda (contents env name)
    (let ((result (scode-eval contents env)))
      (values (or name result)
	      env))))

(define-automatic-property 'exporter '(exports environment)
  #f
  (lambda (exports environment)
    (let ((export-alist
	   (map (lambda (export)
		  (cons (library-export-to export)
			(environment-safe-lookup environment
						 (library-export-from export))))
		exports)))
      (lambda (name)
	(let ((p (assq name export-alist)))
	  (if (not p)
	      (error "Not an exported name:" name))
	  (cdr p))))))

(define (preregister-standard-libraries!)
  (parameterize ((param:hide-notifications? #t))
    (let ((pn (system-library-directory-pathname "libraries")))
      (if pn
	  (find-scheme-libraries! pn)))
    (let ((pn (init-file-specifier->pathname '("libraries"))))
      (if (file-directory? pn)
	  (find-scheme-libraries! (pathname-as-directory pn))))))

(add-boot-init!
 (lambda ()
   (add-event-receiver! event:after-restart preregister-standard-libraries!)))

(define (find-scheme-libraries! pathname)
  (preregister-libraries! pathname (current-library-db)))

(define (preregister-libraries! pathname db)
  (let ((pattern (get-directory-read-pattern pathname)))
    (if pattern
	(let ((root (directory-pathname pattern)))
	  (let loop ((pattern pattern))
	    (receive (files subdirs)
		(find-matching-files scheme-pathname? pattern)
	      (for-each (lambda (group)
			  (preregister-scheme-file! group root db))
			(group-scheme-files files))
	      (for-each (lambda (subdir)
			  (loop (pathname-as-directory subdir)))
			subdirs)))))))

(define (get-directory-read-pattern pathname)
  (case (file-type-direct pathname)
    ((regular)
     (and (or (scheme-pathname? pathname)
	      (not (pathname-type pathname)))
	  (pathname-new-type pathname 'wild)))
    ((directory) (pathname-as-directory pathname))
    (else #f)))

(define (find-matching-files predicate pattern)
  (let loop
      ((pathnames (directory-read pattern #f))
       (files '())
       (directories '()))
    (if (pair? pathnames)
	(let ((pathname (car pathnames))
	      (pathnames (cdr pathnames)))
	  (case (file-type-direct pathname)
	    ((regular)
	     (loop pathnames
		   (if (predicate pathname)
		       (cons pathname files)
		       files)
		   directories))
	    ((directory)
	     (loop pathnames
		   files
		   (if (member (file-namestring pathname) '("." ".."))
		       directories
		       (cons pathname directories))))
	    (else
	     (loop pathnames files directories))))
	(values files directories))))

(define (scheme-pathname? pathname)
  (member (pathname-type pathname) '("scm" "bin" "com")))

(define (group-scheme-files files)
  (let ((table (make-string-hash-table)))
    (for-each (lambda (file)
		(hash-table-update! table
				    (pathname-name file)
				    (lambda (files) (cons file files))
				    (lambda () '())))
	      files)
    (hash-table-values table)))

(define (preregister-scheme-file! file-group root db)

  (define (find-file type)
    (find (lambda (file)
	    (string=? type (pathname-type file)))
	  file-group))

  (define (handle-compiled compiled)
    (let ((scode (fasload compiled)))
      (if (r7rs-scode-file? scode)
	  (finish (map (lambda (scode-library)
			 (scode-library->library scode-library
						 (->namestring compiled)))
		       (r7rs-scode-file-libraries scode))
		  compiled))))

  (define (handle-source source)
    (let ((parsed (read-r7rs-source source)))
      (if parsed
	  (finish (r7rs-source-libraries parsed) source))))

  (define (finish libraries filename)
    (let ((display-filename (enough-namestring filename root)))
      ;; Remove any previously registed libraries from this file.
      (for-each (let ((no-type (pathname-new-type filename #f)))
		  (lambda (library)
		    (if (let ((filename* (library-filename library)))
			  (and filename*
			       (pathname=? (pathname-new-type filename* #f)
					   no-type)))
			(with-notification
			    (lambda (port)
			      (write-string "Deregistering library " port)
			      (write (library-name library) port)
			      (write-string " from \"" port)
			      (write-string
			       (enough-namestring (library-filename library)
						  root)
			       port)
			      (write-string "\"" port))
			  (lambda ()
			    (db 'delete! (library-name library)))))))
		(db 'get-all))
      ;; Now register the current file contents.
      (for-each (lambda (library)
		  (with-notification
		      (lambda (port)
			(write-string "Registering library " port)
			(write (library-name library) port)
			(write-string " from \"" port)
			(write-string display-filename port)
			(write-string "\"" port))
		    (lambda ()
		      (preregister-library! library db))))
		libraries)))

  (let ((compiled (or (find-file "com") (find-file "bin")))
	(source (find-file "scm")))
    (if compiled
	(handle-compiled compiled)
	(begin
	  (if (not source)
	      (error "No scheme files:" file-group))
	  (handle-source source)))))

(define (maybe-load-libraries! imports db)
  (let ((libraries (dependency-libraries imports db)))
    (if (any library-preregistered? libraries)
	(for-each (lambda (library)
		    (load (library-filename library)))
		  (reverse
		   (filter library-preregistered?
			   ((compute-dependency-graph libraries db)
			    'topological-sort)))))))

(define (dependency-libraries imports db)
  (let ((names
	 (delete-duplicates! (map library-import-from-library imports)
			     equal?)))
    (let ((unregistered
	   (remove (lambda (name)
		     (registered-library? name db))
		   names)))
      (if (pair? unregistered)
	  (error "Unknown libraries:" unregistered)))
    (map (lambda (name) (registered-library name db))
	 names)))

(define (compute-dependency-graph libraries db)
  (let ((table (make-key-weak-eq-hash-table)))

    (define (trace library)
      (if (not (hash-table-exists? table library))
	  (let ((deps
		 (dependency-libraries (library-imports library)
				       db)))
	    (hash-table-set! table library deps)
	    (for-each trace deps))))

    (for-each trace libraries)
    (make-digraph (hash-table-keys table)
		  (lambda (library) (hash-table-ref table library)))))